﻿namespace PracticalAgileCrypto;

using System;
using System.Text;
using System.IO;
using System.Security.Cryptography;


public class AgileCrypto
{
    // Salt-Größe ist die Anzahl der Byte, nicht der Bits 
    private const int Saltsize = 128 / 8;

    // Für Testzwecke - 4 Versionen
    public enum Version
    {
        Version1 = 1,
        Version2 = 2,
        Version3 = 3,
        Version4 = 4,
        VersionLatest = Version4
    };

    // 'Delim' wird verwendet, um die Elenente im sich ergebenden String
    // voneinander zu trennen
    private const char Delim = '|';

    private Version _ver;
    private SymmetricAlgorithm? _symCrypto;
    private HMAC? _hMac;
    private DeriveBytes? _keyDerivation;
    private int _iterationCount;
    private byte[] _salt;
    private byte[]? _keyMaterial;

    public AgileCrypto(Version ver)
    {
        _ver = ver;
        _salt = RandomNumberGenerator.GetBytes(Saltsize);
    }

    public AgileCrypto()
        : this(Version.VersionLatest)
    {
    }

    /// <summary>
    /// Erstellt anhand der Versionsnummer die internen Krypto-Klassen
    /// </summary>
    /// <exception cref="ArgumentException"></exception>
    private void BuildCryptoObjects(string pwd)
    {
        _keyMaterial = Encoding.ASCII.GetBytes(pwd);

        switch (_ver)
        {
            case Version.Version1:
                _symCrypto = CreateSymmetricAlgorithm("DES");
                _symCrypto.Mode = CipherMode.ECB;
                _symCrypto.Padding = PaddingMode.PKCS7;
                _hMac = new HMACMD5();
                _iterationCount = 100;
                _keyDerivation = new Rfc2898DeriveBytes(_keyMaterial, _salt, _iterationCount, HashAlgorithmName.SHA1);
                break;

            case Version.Version2:
                _symCrypto = CreateSymmetricAlgorithm("TripleDes");
                _symCrypto.KeySize = 128;
                _symCrypto.Mode = CipherMode.CBC;
                _symCrypto.Padding = PaddingMode.PKCS7;
                _hMac = new HMACMD5();
                _iterationCount = 1000;
                _keyDerivation = new Rfc2898DeriveBytes(_keyMaterial, _salt, _iterationCount, HashAlgorithmName.SHA1);
                break;

            case Version.Version3:
                _symCrypto = CreateSymmetricAlgorithm("AesManaged");
                _symCrypto.KeySize = 128;
                _symCrypto.Mode = CipherMode.CBC;
                _symCrypto.Padding = PaddingMode.PKCS7;
                _hMac = new HMACSHA1();
                _iterationCount = 4000;
                _keyDerivation = new Rfc2898DeriveBytes(_keyMaterial, _salt, _iterationCount, HashAlgorithmName.SHA256);
                break;

            case Version.Version4:
                _symCrypto = CreateSymmetricAlgorithm("AesManaged");
                _symCrypto.KeySize = 256;
                _symCrypto.Mode = CipherMode.CBC;
                _symCrypto.Padding = PaddingMode.ANSIX923;
                _hMac = new HMACSHA256();
                _iterationCount = 20000;
                _keyDerivation = new Rfc2898DeriveBytes(_keyMaterial, _salt, _iterationCount, HashAlgorithmName.SHA512);
                break;

            default:
                throw new ArgumentException("Ungültige Krypto-Vversion.");
        }
    }

    /// <summary>
    /// Methode, um Klartext zu verschlüsseln, und dann daraus MAC zu erstellen
    /// </summary>
    /// <param name="pwd"></param>
    /// <param name="plaintext"></param>
    /// <returns>Base64-codierter String der enthält: Versionsinfo, IV, PBKDF# usw.</returns>
    public string Protect(string pwd, string plaintext)
    {
        BuildCryptoObjects(pwd);

        var sb = new StringBuilder();

        if (_symCrypto is not null && _keyDerivation is not null && _hMac is not null)
        {
            byte[] encrypted = EncryptThePlaintext(plaintext);

            sb.Append((int)_ver)
                .Append(Delim)
                .Append(Convert.ToBase64String(_symCrypto.IV))
                .Append(Delim)
                .Append(Convert.ToBase64String(_salt))
                .Append(Delim)
                .Append(Convert.ToBase64String(encrypted))
                .Append(Delim);

            // Erstelle jetzt einen HMAC über alle obigen Daten
            // einschl. Versionsnummer, IV, salt und Ciphertext
            // bis auf den Chiffretext ist alles Klartext, wir schützen die
            // Daten vor Manipulation

            // Für diese Arbeit einen neuen Schlüssel erstellen
            _hMac.Key = _keyDerivation.GetBytes(_hMac.HashSize);
            _hMac.ComputeHash(Encoding.UTF8.GetBytes(sb.ToString()));
            sb.Append(Convert.ToBase64String(_hMac.Hash));

            return sb.ToString();
        }
        else
        {
            throw new ArgumentNullException("Crypto und/oder Derivation und/oder MAC ist NULL");
        }
    }

    /// <summary>
    /// Method to verify the MAC and then decrypt a protected blob if the HMAC is ok.
    /// </summary>
    /// <param name="pwd"></param>
    /// <param name="protectedBlob"></param>
    /// <returns>The plaintext string</returns>
    /// <exception cref="ArgumentException"></exception>
    public string Unprotect(string pwd, string protectedBlob)
    {
        if (string.IsNullOrWhiteSpace(protectedBlob))
            throw new ArgumentException($"'{nameof(protectedBlob)}' kann nicht NULL oder leer sein.", nameof(protectedBlob));

        BuildCryptoObjects(pwd);

        if (_symCrypto is not null && _keyDerivation is not null && _hMac is not null)
        {

            // Die einzelnen Elemente aus dem geschützten Blob extrahieren
            // 0: Version
            // 1: IV
            // 2: Salt
            // 3: Ciphertext
            // 4: MAC
            const int version = 0, initvect = 1, salt = 2, ciphertext = 3, mac = 4;
            string[] elements = protectedBlob.Split(new char[] { Delim });

            // Versionsnummer abrufen
            int.TryParse(elements[version], out int ver);
            _ver = (Version)ver;

            // IV/Salt/Chiffretext abrufen 
            byte[] iv = Convert.FromBase64String(elements[initvect]);
            _salt = Convert.FromBase64String(elements[salt]);
            byte[] ctext = Convert.FromBase64String(elements[ciphertext]);

            // Wir haben nun alle Daten, um die Krypto-Algorithmen zu erstellen
            BuildCryptoObjects(pwd);

            _symCrypto.Key = _keyDerivation.GetBytes(_symCrypto.KeySize >> 3);
            _symCrypto.IV = iv;

            // Bevor wir den Chiffretext entschlüssen, müssen wir den MAC überprüfen
            if (string.IsNullOrWhiteSpace(elements[mac]))
                throw new ArgumentException($"'{nameof(protectedBlob)}' MAC fehlt.", nameof(protectedBlob));

            // Überprüfen des HMAC, dies funktioniert wie folgt:
            // 1) den HMAC aus dem geschützten Blob entnehmen
            // 2) mit dem oben erzeugten String einen HMAC erstellen 
            // 3) den HMAC im geschützen Blob und den erstellen HMAC vergleichen
            string blobLessMac = protectedBlob.Substring(0, protectedBlob.LastIndexOf(elements[mac]));
            _hMac.Key = _keyDerivation.GetBytes(_hMac.HashSize);
            byte[] result = _hMac.ComputeHash(Encoding.UTF8.GetBytes(blobLessMac));
            if (string.Compare(elements[mac], Convert.ToBase64String(result), true) != 0)
                throw new ArgumentException($"'{nameof(protectedBlob)}' MAC ungültig.", nameof(protectedBlob));

            string plaintext = DecryptThePlainText(ctext);

            return plaintext;
        }
        else
        {
            throw new ArgumentNullException("Crypto und/oder Derivation und/oder MAC ist NULL");
        }
    }

    private static SymmetricAlgorithm CreateSymmetricAlgorithm(string name)
    {
        SymmetricAlgorithm? result = SymmetricAlgorithm.Create(name);
        if (result is null)
            throw new InvalidOperationException($"Der symmetrische Algorithmus {name} kann nicht erstellt werden.");

        return result;
    }

    private byte[] EncryptThePlaintext(string plaintext)
    {
        if (_symCrypto is null || _keyDerivation is null)
            throw new ArgumentNullException("Symmetrischer und Schlüsselableitungsalgorithmus dürfen nicht NULL sein.");

        _symCrypto.GenerateIV();
        _symCrypto.Key = _keyDerivation.GetBytes(_symCrypto.KeySize >> 3);

        byte[] encrypted;

        ICryptoTransform encryptor = _symCrypto.CreateEncryptor();
        using var msEncrypt = new MemoryStream();
        using var csEncrypt = new CryptoStream(msEncrypt, encryptor, CryptoStreamMode.Write);
        using var swEncrypt = new StreamWriter(csEncrypt);
        swEncrypt.Write(plaintext);
        swEncrypt.Close();
        encrypted = msEncrypt.ToArray();

        return encrypted;
    }

    private string DecryptThePlainText(byte[] ctext)
    {
        if (_symCrypto == null)
            throw new ArgumentNullException("Symmetrischer Algorithmus darf nicht NULL sein.");

        string plaintext;
        ICryptoTransform decryptor = _symCrypto.CreateDecryptor();
        using var msDecrypt = new MemoryStream(ctext);
        using var csDecrypt = new CryptoStream(msDecrypt, decryptor, CryptoStreamMode.Read);
        using var srDecrypt = new StreamReader(csDecrypt);
        plaintext = srDecrypt.ReadToEnd();

        return plaintext;
    }
}

static class Program
{
    static void Main()
    {
        // Ein paar Tests ... ja, pwd und plaintext befinden sich im Code!
        const string pwd = "SSsshh!!";
        const string plaintext = "Hello, World!";

        string c1 = new AgileCrypto(AgileCrypto.Version.Version1).Protect(pwd, plaintext);
        string p1 = new AgileCrypto().Unprotect(pwd, c1);
        Console.WriteLine($"P1 {p1 == plaintext}");

        string c2 = new AgileCrypto(AgileCrypto.Version.Version2).Protect(pwd, plaintext);
        string p2 = new AgileCrypto().Unprotect(pwd, c2);
        Console.WriteLine($"P2 {p2 == plaintext}");

        string c3 = new AgileCrypto(AgileCrypto.Version.Version3).Protect(pwd, plaintext);
        string p3 = new AgileCrypto().Unprotect(pwd, c3);
        Console.WriteLine($"P3 {p3 == plaintext}");

        string c4 = new AgileCrypto(AgileCrypto.Version.Version4).Protect(pwd, plaintext);
        string p4 = new AgileCrypto().Unprotect(pwd, c4);
        Console.WriteLine($"P4 {p4 == plaintext}");

        string c5 = new AgileCrypto(AgileCrypto.Version.Version4).Protect(pwd, plaintext);
        string p5 = new AgileCrypto().Unprotect(pwd, c5);
        Console.WriteLine($"P5 {p5 == plaintext}");

        // Zwei Klartexte die mit dem gleichen Algorithmus und Schlüssel
        // verschlüsselt werden, sollten unterschiedliche Chiffretexte ergeben 
        // da IV und Salt immer anders sind
        Console.WriteLine($"C4 != C5 {c4 != c5}");

        // Ein einzelnes Objekt zum Verschlüsseln und Entschlüsseln testen 
        var ac = new AgileCrypto();
        string c6 = ac.Protect(pwd, plaintext);
        string c7 = ac.Protect(pwd, plaintext);
        string p6 = ac.Unprotect(pwd, c6);
        string p7 = ac.Unprotect(pwd, c7);
        Console.WriteLine($"P6 && P7 {p6 == plaintext && p7 == plaintext}");
    }
}